////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// BeDVI                                                                                                          //
// by Achim Blumensath                                                                                            //
// blume@corona.oche.de                                                                                           //
//                                                                                                                //
// This program is free software! It may be distributed according to the GNU Public License (see COPYING).        //
//                                                                                                                //
// $Id: DVI.cc,v 1.3 1996/12/26 15:34:26 achim Exp achim $
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <InterfaceKit.h>
#include <StorageKit.h>
#include "debug.h"
#include "defines.h"
extern "C" {
#include "kpathsea/magstep.h"
};
#include "BeDVI.h"
#include "DVI-View.h"
#include "DVI.h"
#include "FontList.h"
#include "Font.h"

static const int NoMagStep = -29999;
static const int NoBuild   =  29999;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// u_long DrawInfo::ReadInt(BFile &f, size_t Size)                                                                //
//                                                                                                                //
// Reads an integer of `Size' bytes from `f' or from a buffer.                                                    //
//                                                                                                                //
// BFile  &f                            file                                                                      //
// size_t Size                          size of the integer in bytes                                              //
//                                                                                                                //
// Result:                              read integer                                                              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

u_long DrawInfo::ReadInt(BFile &f, size_t Size)
{
  u_long x = 0;

  assert(Size > 0 && Size <= sizeof(u_long));

  if(BufferPos == NULL)
    f.Read((u_char *)&x + sizeof(u_long) - Size, Size);
  else
  {
    memcpy((u_char *)&x + sizeof(u_long) - Size, BufferPos, Size);
    BufferPos += Size;
  }

  return x;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// long DrawInfo::ReadSInt(BFile &f, size_t Size)                                                                 //
//                                                                                                                //
// Reads a signed integer of `Size' bytes from `f' or from a buffer.                                              //
//                                                                                                                //
// BFile  &f                            file                                                                      //
// size_t Size                          size of the integer in bytes                                              //
//                                                                                                                //
// Result:                              read integer                                                              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

long DrawInfo::ReadSInt(BFile &f, size_t Size)
{
  long x = 0;

  assert(Size > 0 && Size <= sizeof(long));

  if(BufferPos == NULL)
    f.Read((u_char *)&x + sizeof(long) - 1, 1);
  else
  {
    memcpy((u_char *)&x + sizeof(long) - 1, BufferPos, 1);
    BufferPos += 1;
  }

  if(x & 0x80)
    x |= 0xffffff00;

  x <<= (Size - 1) * 8;

  if(Size > 1)
    if(BufferPos == NULL)
      f.Read((u_char *)&x + sizeof(long) - Size + 1, Size - 1);
    else
    {
      memcpy((u_char *)&x + sizeof(long) - Size + 1, BufferPos, Size - 1);
      BufferPos += Size - 1;
    }

  return x;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DisplayInfo::SetResolution(u_int ppi, u_int sf)                                                           //
//                                                                                                                //
// Sets the resolution information.                                                                               //
//                                                                                                                //
// u_int ppi                            dpi value (0: don't change old value)                                     //
// u_int sf                             factor to shrink document (0: don't change old value)                     //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DisplayInfo::SetResolution(u_int ppi, u_int sf)
{
  int i;

  D(bug("DisplayInfo::SetResolution(%u,%u)\n", ppi, sf));

  if(ppi != 0)
  {
    // search for mode name

    for(i = NumModes - 1; i > 0; i--)
      if(ModeDPI[i] <= ppi)
        break;

    Mode          = ModeNames[i];
    PixelsPerInch = ppi;
  }
  if(sf != 0)
    ShrinkFactor  = sf;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// DVI::DVI(record_ref File, DisplayInfo *dsp)                                                                    //
//                                                                                                                //
// Initializes a DVI-Document.                                                                                    //
//                                                                                                                //
// record_ref  File                     file of the document                                                      //
// DisplayInfo *dsp                     data about the display (resolution,...)                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

DVI::DVI(record_ref File, DisplayInfo *dsp):
  Name(NULL),
  PageOffset(NULL),
  DspInfo(),
  Magnification(1000),
  DimConvert(1.0),
  OffsetX(dsp->PixelsPerInch),
  OffsetY(dsp->PixelsPerInch),
  BorderLine(false)
{
  D(bug("DVI::DVI()\n"));

  assert(dsp != NULL);

  DspInfo = *dsp;

  if(!does_ref_conform(File, "File"))
    return;

  if(DVIFile.SetRef(File) != B_NO_ERROR)
    return;

  if(DVIFile.Open(B_READ_ONLY) != B_NO_ERROR)
    return;

  if(!Reload())
  {
    if(Name)
      delete [] Name;
    if(PageOffset)
      delete [] PageOffset;

    Name       = NULL;
    PageOffset = NULL;

    DVIFile.Close();
    return;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// DVI::~DVI()                                                                                                    //
//                                                                                                                //
// Deletes a DVI-Document.                                                                                        //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

DVI::~DVI()
{
  D(bug("DVI::~DVI()\n"));

  if(Name)
    delete [] Name;
  if(PageOffset)
    delete [] PageOffset;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// int DVI::MagStepValue(float &mag)                                                                              //
//                                                                                                                //
// Adjusts a magnification value.                                                                                 //
//                                                                                                                //
// float &mag                           magnification                                                             //
//                                                                                                                //
// Result:                              adjusted value                                                            //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int DVI::MagStepValue(float &mag)
{
  int   Result;
  u_int dpi;

  dpi = kpse_magstep_fix((u_int)mag, DspInfo.PixelsPerInch, &Result);
  mag = (float)dpi;

  return Result ? Result : NoMagStep;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool DVI::Reload()                                                                                             //
//                                                                                                                //
// Reads general data of a DVI file.                                                                              //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool DVI::Reload()
{
  static const int BufferLen = 512;

  long   Numerator;
  long   Denominator;
  int    len;
  long   LastPageOffset;
  long   pos;
  u_char *p;
  u_char Buffer[BufferLen];
  u_char x;
  u_char Command;
  int    i;

  D(bug("DVI::Reload()\n"));

  DVIFile.Seek(0, B_SEEK_TOP);

  if(ReadInt(DVIFile, 1) != Preamble)
    return false;
  if(ReadInt(DVIFile, 1) != 2)
    return false;

  Numerator     = ReadInt(DVIFile, 4);
  Denominator   = ReadInt(DVIFile, 4);
  Magnification = ReadInt(DVIFile, 4);
  DimConvert    = (((double)Numerator * Magnification) / ((double)Denominator * 1000.0));
  DimConvert    = DimConvert * (((long)DspInfo.PixelsPerInch) << 16) / 254000;
  TPicConvert   = DspInfo.PixelsPerInch * Magnification / 1000000.0;

  len = ReadInt(DVIFile, 1);

  if(!(Name = new char[len + 1]))
    return false;

  DVIFile.Read(Name, len);
  Name[len] = 0;

  D(bug("Numerator = %ld\nDenominator = %ld\nMagnification = %ld\nDimConvert = %g\nTPicConvert = %g\n",
        Numerator, Denominator, Magnification, DimConvert, TPicConvert));

  pos = DVIFile.Size();

  if(pos > BufferLen)
    pos -= BufferLen;
  else
    pos = 0;

  DVIFile.Seek(pos, B_SEEK_TOP);

  p = &Buffer[DVIFile.Read(&Buffer, BufferLen)];

  do
  {
    while(p > Buffer && *--p != Trailer)
      ;

    if(p <= Buffer)
      return false;
  }
  while(p[-1] != Trailer || p[-2] != Trailer || p[-3] != Trailer);

  pos += p - Buffer;

  for(x = *p; x == Trailer; x = ReadInt(DVIFile, 1))
    DVIFile.Seek(--pos, B_SEEK_TOP);

  if(x != 2)
    return false;

  DVIFile.Seek(pos - 4, B_SEEK_TOP);
  DVIFile.Seek((long)ReadInt(DVIFile, 4), B_SEEK_TOP);

  if(ReadInt(DVIFile, 1) != Postamble)
    return false;

  LastPageOffset = ReadInt(DVIFile, 4);

  if(ReadInt(DVIFile, 4) != Numerator || ReadInt(DVIFile, 4) != Denominator || ReadInt(DVIFile, 4) != Magnification)
    return false;

  UnshrunkPageHeight = ((long)((long)ReadInt(DVIFile, 4) * DimConvert) >> 16) + 2 * DspInfo.PixelsPerInch;
  UnshrunkPageWidth  = ((long)((long)ReadInt(DVIFile, 4) * DimConvert) >> 16) + 2 * DspInfo.PixelsPerInch;

  D(bug("unshrunk page height = %ld\nunshrunk page width = %ld\n", UnshrunkPageHeight, UnshrunkPageWidth));

  ReadInt(DVIFile, 2);

  NumPages = ReadInt(DVIFile, 2);

  Fonts.FreeFonts();

  for(Command = ReadInt(DVIFile, 1); Command >= FontDef1 && Command <= FontDef4; Command = ReadInt(DVIFile, 1))
    if(!Fonts.LoadFont(this, DVIFile, Command))
      ((ViewApplication *)be_app)->DisplayError("Font not found!");

  FontData.FreeUnusedFonts();

  if(Command != PostPost)
    return false;

  if(PageOffset)
    delete [] PageOffset;

  if(!(PageOffset = new u_long[NumPages]))
    return false;

  PageOffset[NumPages - 1] = LastPageOffset;
  DVIFile.Seek(LastPageOffset, B_SEEK_TOP);

  for(i = NumPages - 2; i >= 0; i--)
  {
    DVIFile.Seek(41, B_SEEK_MIDDLE);
    DVIFile.Seek(PageOffset[i] = ReadInt(DVIFile, 4), B_SEEK_TOP);
  }

  PageWidth  = (UnshrunkPageWidth  + DspInfo.ShrinkFactor - 1) / DspInfo.ShrinkFactor + 2;
  PageHeight = (UnshrunkPageHeight + DspInfo.ShrinkFactor - 1) / DspInfo.ShrinkFactor + 2;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::SetDspInfo(DisplayInfo *dsp)                                                                         //
//                                                                                                                //
// Sets display data.                                                                                             //
//                                                                                                                //
// DisplayInfo *dsp                     display information                                                       //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::SetDspInfo(DisplayInfo *dsp)
{
  assert(dsp != NULL);

  DspInfo = *dsp;

  OffsetX = DspInfo.PixelsPerInch;
  OffsetY = DspInfo.PixelsPerInch;

  Reload();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::Special(long len)                                                                                    //
//                                                                                                                //
// Processes `special' commands (currently not implemented).                                                      //
//                                                                                                                //
// long len                             length of the command                                                     //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::Special(long len)
{
  static char *Cmd   = NULL;
  static long CmdLen = -1;
         char *p;

  if(CmdLen < len)
  {
    if(Cmd)
      delete [] Cmd;

    if(Cmd = new char[len + 1])
      CmdLen = len;
    else
      CmdLen = -1;
  }
  if(Cmd)
  {
    DVIFile.Read(Cmd, len);
    Cmd[len] = 0;
  }
  else
    DVIFile.Seek(len, B_SEEK_MIDDLE);

  D(bug("special: `%s'\n", Cmd));

  // do something!?
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::ChangeFont(u_long n)                                                                                 //
//                                                                                                                //
// Changes the current font.                                                                                      //
//                                                                                                                //
// u_long n                             number of the new font                                                    //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::ChangeFont(DrawInfo *di, u_long n)
{
  if(n < Fonts.TableLen)
    if((di->CurFont = Fonts.Table[n]) != NULL)
    {
      di->MaxChar = di->CurFont->MaxChar;
      di->SetChar = di->CurFont->SetChar;
    }
    else
    {
      di->MaxChar = 255;
      di->SetChar = DVIView::SetNoChar;
    }
  else
    ((ViewApplication *)be_app)->DisplayError("Non-existent font number!");
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::Draw(BView *vw, BRect r, u_int PageNo)                                                               //
//                                                                                                                //
// Draws a DVI-Document.                                                                                          //
//                                                                                                                //
// BView *vw                            view the document should be displayed in                                  //
// BRect r                              rectangle to be drawn                                                     //
// u_int PageNo                         page to be displayed                                                      //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::Draw(BView *vw, u_int PageNo)
{
  DrawInfo di;

  assert(vw != NULL);

  vw->SetDrawingMode(B_OP_COPY);

  vw->FillRect(vw->Bounds(), B_SOLID_LOW);

  if(((ViewApplication *)be_app)->AntiAliasedDisplay && DspInfo.ShrinkFactor > 1)
    vw->SetDrawingMode(B_OP_MIN);
  else
    vw->SetDrawingMode(B_OP_OVER);

  DVIFile.Seek(PageOffset[PageNo - 1], B_SEEK_TOP);

  di.vw           = (DVIView *)vw;
  di.DspInfo      = &DspInfo;
  di.DimConvert   = DimConvert;
  di.Virtual      = NULL;
  di.CurFont      = NULL;
  di.SetChar      = DVIView::SetNoChar;
  di.BufferPos    = NULL;
  di.BufferEnd    = NULL;

  bzero(&di.Data, sizeof(di.Data));

  DrawPart(&di);

  if(BorderLine)
  {
    BRect   r;
    pattern Dash = {0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f};

    vw->SetDrawingMode(B_OP_COPY);

    r.Set(DspInfo.PixelsPerInch / DspInfo.ShrinkFactor,
          DspInfo.PixelsPerInch / DspInfo.ShrinkFactor,
          PageWidth  - DspInfo.PixelsPerInch / DspInfo.ShrinkFactor,
          PageHeight - DspInfo.PixelsPerInch / DspInfo.ShrinkFactor);

    vw->StrokeLine(BPoint(r.left,  0.0),      BPoint(r.left,        PageHeight - 1), Dash);
    vw->StrokeLine(BPoint(r.right, 0.0),      BPoint(r.right,       PageHeight - 1), Dash);
    vw->StrokeLine(BPoint(0.0,     r.top),    BPoint(PageWidth - 1, r.top),          Dash);
    vw->StrokeLine(BPoint(0.0,     r.bottom), BPoint(PageWidth - 1, r.bottom),       Dash);
  }

  vw->Sync();
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void DVI::DrawPart(DrawInfo *di)                                                                               //
//                                                                                                                //
// Draws a part of a document.                                                                                    //
//                                                                                                                //
// DrawInfo *di                         drawing information                                                       //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void DVI::DrawPart(DrawInfo *di)
{
  u_char c;

  while(true)
  {
    c = di->ReadInt(DVIFile, 1);

    if(c <= (u_char)(SetChar0 + 127))
      di->Data.Horiz += (*di->SetChar)(di, c);

    else if(FontNum0 <= c && c <= (u_char)(FontNum0 + 63))
      ChangeFont(di, (u_long)(c - FontNum0));

    else
    {
      long a, b;

      switch(c)
      {
        case Set1:
          di->Data.Horiz += (*di->SetChar)(di, di->ReadInt(DVIFile, 1));
          break;

        case Put1:
          (*di->SetChar)(di, di->ReadInt(DVIFile, 1));
          break;

        case SetRule:
          a = di->ReadInt(DVIFile, 4) * DimConvert;
          b = di->ReadInt(DVIFile, 4) * DimConvert;

          if(a > 0 && b > 0)
            DVIView::DrawRule(di, DspInfo.PixelConv(di->Data.Horiz), di->Data.PixelV - DspInfo.ToPixel(a) + 1,
                                  DspInfo.ToPixel(b),                DspInfo.ToPixel(a));

          di->Data.Horiz += b;
          break;

        case PutRule:
          a = di->ReadInt(DVIFile, 4) * DimConvert;
          b = di->ReadInt(DVIFile, 4) * DimConvert;

          if(a > 0 && b > 0)
            DVIView::DrawRule(di, DspInfo.PixelConv(di->Data.Horiz), di->Data.PixelV - DspInfo.ToPixel(a) + 1,
                                  DspInfo.ToPixel(b), DspInfo.ToPixel(a));
          break;

        case NOP:
          break;

        case BeginOP:
          DVIFile.Seek(44, B_SEEK_MIDDLE);

          di->Data.Horiz  = OffsetX << 16;
          di->Data.Vert   = OffsetY << 16;
          di->Data.PixelV = DspInfo.PixelConv(di->Data.Vert);
          di->Data.w      = 0;
          di->Data.x      = 0;
          di->Data.y      = 0;
          di->Data.z      = 0;

          break;

        case EndOP:
          if(!di->FrameList.Empty())
            ((ViewApplication *)be_app)->DisplayError("Stack not empty at EOP!");
          return;

        case Push:
        {
          Frame *f;

          if(!di->FreeFrames.Empty())
            f = (Frame *)di->FreeFrames.RemHead();
          else
            f = new Frame();

          if(!f)
          {
            ((ViewApplication *)be_app)->DisplayError("Not enough memory!");
            return;
          }
          f->Data = di->Data;
          di->FrameList.AddTail(f);
          break;
        }
        case Pop:
        {
          Frame *f;

          if(di->FrameList.Empty())
          {
            ((ViewApplication *)be_app)->DisplayError("Stack empty at Pop!");
            return;
          }

          f = (Frame *)di->FrameList.RemTail();

          di->Data = f->Data;
          di->FreeFrames.AddTail(f);
          break;
        }
        case Right1:
        case Right2:
        case Right3:
        case Right4:
          di->Data.Horiz += di->ReadSInt(DVIFile, c - Right1 + 1) * DimConvert;
          break;

        case W1:
        case W2:
        case W3:
        case W4:
          di->Data.w      = di->ReadSInt(DVIFile, c - W0) * DimConvert;
        case W0:
          di->Data.Horiz += di->Data.w;
          break;

        case X1:
        case X2:
        case X3:
        case X4:
          di->Data.x      = di->ReadSInt(DVIFile, c - X0) * DimConvert;
        case X0:
          di->Data.Horiz += di->Data.x;
          break;

        case Down1:
        case Down2:
        case Down3:
        case Down4:
          di->Data.Vert  += di->ReadSInt(DVIFile, c - Down1 + 1) * DimConvert;
          di->Data.PixelV = DspInfo.PixelConv(di->Data.Vert);
          break;

        case Y1:
        case Y2:
        case Y3:
        case Y4:
          di->Data.y      = di->ReadSInt(DVIFile, c - Y0) * DimConvert;
        case Y0:
          di->Data.Vert  += di->Data.y;
          di->Data.PixelV = DspInfo.PixelConv(di->Data.Vert);
          break;

        case Z1:
        case Z2:
        case Z3:
        case Z4:
          di->Data.z      = di->ReadSInt(DVIFile, c - Z0) * DimConvert;
        case Z0:
          di->Data.Vert  += di->Data.z;
          di->Data.PixelV = DspInfo.PixelConv(di->Data.Vert);
          break;

        case Font1:
        case Font2:
        case Font3:
        case Font4:
          ChangeFont(di, di->ReadInt(DVIFile, c - Font1 + 1));
          break;

        case XXX1:
        case XXX2:
        case XXX3:
        case XXX4:
          a = di->ReadInt(DVIFile, c - XXX1 + 1);

          if(a > 0)
            Special(a);
          break;

        case FontDef1:
        case FontDef2:
        case FontDef3:
        case FontDef4:
          DVIFile.Seek(13 + c - FontDef1, B_SEEK_MIDDLE);
          DVIFile.Seek(di->ReadInt(DVIFile, 1) + di->ReadInt(DVIFile, 1), B_SEEK_MIDDLE);
          break;

        default:
          D(bug("invalid operand: 0x%02x!\n", (u_int)c));

          ((ViewApplication *)be_app)->DisplayError("Invalid operand!");
          return;
      }
    }
  }
}
