////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// BeDVI                                                                                                          //
// by Achim Blumensath                                                                                            //
// blume@corona.oche.de                                                                                           //
//                                                                                                                //
// This program is free software! It may be distributed according to the GNU Public License (see COPYING).        //
//                                                                                                                //
// $Id: Font.cc,v 1.3 1996/12/26 15:34:26 achim Exp achim $
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "defines.h"
#include <stdio.h>
#include <string.h>
extern "C" {
#include "kpathsea/tex-glyph.h"
};
#include "debug.h"
#include "BeDVI.h"
#include "DVI-View.h"
#include "Font.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// Font::Font(DVI *doc, const char *name, float size, long chksum, int magstep, double dimconvert)                //
//                                                                                                                //
// Initializes a Font.                                                                                            //
//                                                                                                                //
// DVI        *doc                      document the font appears in                                              //
// const char *name                     name of the font                                                          //
// float      size                      size of the font                                                          //
// long       chksum                    checksum                                                                  //
// int        magstep                   magnification                                                             //
// double     dimconvert                factor to convert dimensions                                              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Font::Font(DVI *doc, const char *name, float size, long chksum, int magstep, double dimconvert):
  Name(NULL),
  Size(size),
  ChkSum(chksum),
  MagStep(magstep),
  DimConvert(dimconvert),
  UseCount(0),
  Loaded(false),
  Virtual(false)
{
  if(name)
    if(Name = new char[strlen(name) + 1])
      strcpy(Name, name);

  Load(doc);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// Font::~Font()                                                                                                  //
//                                                                                                                //
// Deletes a font.                                                                                                //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Font::~Font()
{
  if(Name)
    delete [] Name;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Font::Open(char *&FontFound, int &SizeFound)                                                              //
//                                                                                                                //
// Opens the file of a font.                                                                                      //
//                                                                                                                //
// char *&FontFound                     used to return the name of the font found                                 //
// int  &SizeFound                      used to return the size of the font found                                 //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Font::Open(char *&FontFound, int &SizeFound)
{
  record_ref file;
  char       *name;

  name = kpse_find_vf(Name);

  if(name)
  {
    SizeFound = Size;
    FontFound = NULL;
  }
  else
  {
    kpse_glyph_file_type FileResult;

    name = kpse_find_glyph(Name, (u_int)(Size + 0.5), kpse_any_glyph_format, &FileResult);

    if(name)
    {
      if(FileResult.source == kpse_glyph_source_fallback)
        FontFound = FileResult.name;
      else
        FontFound = NULL;

      SizeFound = FileResult.dpi;
    }
  }

  if(!name)
    return false;

  get_ref_for_path(name, &file);

  if(File.SetRef(file) != B_NO_ERROR)
    return false;

  if(File.Open(B_READ_ONLY) != B_NO_ERROR)
    return false;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void Font::ReallocFont(u_char num)                                                                             //
//                                                                                                                //
// shrinks the arrays containing font data to their minimum size.                                                 //
//                                                                                                                //
// u_char num                           used length of array                                                      //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Font::ReallocFont(u_char num)
{
  if(Virtual)
  {
    Macro *m;

    if(!(m = new Macro[num]))
      return;

    memcpy(m, Macros, num * sizeof(Macro));

    bzero(Macros, 256 * sizeof(Macro));
    delete [] Macros;

    Macros = m;
  }
  else
  {
    Glyph *g;

    if(!(g = new Glyph[num]))
      return;

    memcpy(g, Glyphs, num * sizeof(Glyph));

    bzero(Glyphs, 256 * sizeof(Glyphs));
    delete [] Glyphs;

    Glyphs = g;
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Font::Load(DVI *doc)                                                                                      //
//                                                                                                                //
// Loads a font.                                                                                                  //
//                                                                                                                //
// DVI        *doc                      document the font appears in                                              //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Font::Load(DVI *doc)
{
  char *FontFound;
  int  SizeFound;
  int  Type;

  if(!Open(FontFound, SizeFound))
    return false;

  if(FontFound)
  {
    if(Name)
      delete [] Name;
    Name = FontFound;
  }

  Size    = SizeFound;
  MaxChar = 255;
  SetChar = DVIView::SetChar;

  Type = ReadInt(File, 2);

  if(Type == PK_Magic)
  {
    if(!ReadPKIndex(doc, this))
      return false;
  }
  else if(Type == GF_Magic)
  {
    if(!ReadGFIndex(doc, this))
      return false;
  }
  else if(Type == VF_Magic)
  {
    if(!ReadVFIndex(doc, this))
      return false;

    SetChar = DVIView::SetVFChar;
  }
  else
    return false;

  if(Virtual)
    while(MaxChar > 0 && Macros[MaxChar].Position == NULL)
      MaxChar--;
  else
    while(MaxChar > 0 && Glyphs[MaxChar].Addr == 0)
      MaxChar--;

  if(MaxChar < 255)
    ReallocFont(MaxChar + 1);

  Loaded = true;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// void Font::FlushShrinkedGlyphes()                                                                              //
//                                                                                                                //
// Frees all bitmaps of shrinked glyphes which is neccessary if the anti aliasing method is changed.              //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Font::FlushShrinkedGlyphes()
{
  int i;

  for(i = 0; i < MaxChar; i++)
    if(Glyphs[i].SBitMap)
    {
      delete Glyphs[i].SBitMap;
      Glyphs[i].SBitMap = NULL;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Glyph::Shrink(int Factor)                                                                                 //
//                                                                                                                //
// shrinks a glyph.                                                                                               //
//                                                                                                                //
// int Factor                           factor the glyph should be shrinked                                       //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Glyph::Shrink(int Factor)
{
  if(((ViewApplication *)be_app)->AntiAliasedDisplay)
    return ShrinkGrey(Factor);
  else
    return ShrinkMonochrome(Factor);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Glyph::ShrinkMonochrome(int Factor)                                                                       //
//                                                                                                                //
// shrinks a glyph without antialiasing.                                                                          //
//                                                                                                                //
// int Factor                           factor the glyph should be shrinked                                       //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Glyph::ShrinkMonochrome(int Factor)
{
  BRect  r;
  int    Width;
  int    Height;
  int    ShrunkBytesWide;
  int    ShrunkHeight;
  int    RowsLeft;
  int    Rows;
  int    InitCols;
  int    ColsLeft;
  int    Cols;
  u_char *OldPtr;
  u_char *NewPtr;
  u_char m;
  u_char *cp;
  int    MinSample = 0.4 * Factor * Factor;

  if(SBitMap)         // already shrunken?
    return true;

  Sx = Ux / Factor;
  r  = UBitMap->Bounds();

  Width  = r.right - r.left + 1;
  Height = r.bottom - r.top + 1;

  r.right  = r.left + Width / Factor;
  r.bottom = r.top + Height / Factor;

  if(!(SBitMap = new BBitmap(r, B_MONOCHROME_1_BIT)))
    return false;

  OldPtr = (u_char *)UBitMap->Bits();
  NewPtr = (u_char *)SBitMap->Bits();

  ShrunkHeight    = r.bottom - r.top + 1;
  ShrunkBytesWide = SBitMap->BytesPerRow();
  RowsLeft        = Height;
  Rows            = Factor;

  bzero(NewPtr, SBitMap->BitsLength());

  while(RowsLeft)
  {
    if(Rows > RowsLeft)
      Rows = RowsLeft;

    ColsLeft = Width;
    m        = 1 << 7;
    cp       = NewPtr;
    Cols     = Factor;

    while(ColsLeft)
    {
      if(Cols > ColsLeft)
        Cols = ColsLeft;

      if(Sample(OldPtr, UBitMap->BytesPerRow(), Width - ColsLeft, Cols, Rows) >= MinSample)
        *cp |= m;

      if(m == 1)
      {
        m = 1 << 7;
        cp++;
      }
      else
        m >>= 1;

      ColsLeft -= Cols;
      Cols      = Factor;
    }
    NewPtr   += ShrunkBytesWide;
    OldPtr   += Rows * UBitMap->BytesPerRow();
    RowsLeft -= Rows;
    Rows      = Factor;
  }

  Sy = Uy / Factor;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// bool Glyph::ShrinkGrey(int Factor)                                                                             //
//                                                                                                                //
// shrinks a glyph with antialiasing.                                                                             //
//                                                                                                                //
// int Factor                           factor the glyph should be shrinked                                       //
//                                                                                                                //
// Result:                              `true' if successful, otherwise `false'                                   //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Glyph::ShrinkGrey(int Factor)
{
  static u_char *ColourTable      = NULL;
  static int    ColourTableFactor = 0;

  BRect  r;
  int    Width;
  int    Height;
  int    ShrunkBytesWide;
  int    ShrunkHeight;
  int    RowsLeft;
  int    Rows;
  int    ColsLeft;
  int    Cols;
  u_char *OldPtr;
  u_char *NewPtr;
  u_char *cp;
  int    Colour;

  if(SBitMap)         // already shrunken?
    return true;

  if(!ColourTable || ColourTableFactor != Factor)
  {
    if(ColourTable)
      delete [] ColourTable;

    if(!(ColourTable = new u_char [Factor * Factor + 1]))
      return false;

    ColourTableFactor = Factor;

    for(Cols = 0; Cols <= Factor * Factor; Cols++)
    {
      Colour = 255 - (510 * Cols + Factor * Factor) / (2 * Factor * Factor);
      ColourTable[Cols] = index_for_color(Colour, Colour, Colour);
    }
  }

  Sx = Ux / Factor;
  r  = UBitMap->Bounds();

  Width  = r.right - r.left + 1;
  Height = r.bottom - r.top + 1;

  r.right  = r.left + Width / Factor;
  r.bottom = r.top + Height / Factor;

  if(!(SBitMap = new BBitmap(r, B_COLOR_8_BIT)))      // should be replaced by B_GRAYSCALE_8_BIT
    return false;

  OldPtr = (u_char *)UBitMap->Bits();
  NewPtr = (u_char *)SBitMap->Bits();

  ShrunkHeight    = r.bottom - r.top + 1;
  ShrunkBytesWide = SBitMap->BytesPerRow();
  RowsLeft        = Height;
  Rows            = Factor;

  memset(NewPtr, 255, SBitMap->BitsLength());

  while(RowsLeft)
  {
    if(Rows > RowsLeft)
      Rows = RowsLeft;

    ColsLeft = Width;
    Cols     = Factor;
    cp       = NewPtr;

    while(ColsLeft)
    {
      if(Cols > ColsLeft)
        Cols = ColsLeft;

      *cp++ = ColourTable[Sample(OldPtr, UBitMap->BytesPerRow(), Width - ColsLeft, Cols, Rows)];

      ColsLeft -= Cols;
      Cols      = Factor;
    }
    NewPtr   += ShrunkBytesWide;
    OldPtr   += Rows * UBitMap->BytesPerRow();
    RowsLeft -= Rows;
    Rows      = Factor;
  }

  Sy = Uy / Factor;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                //
// int Glyph::Sample(u_long *Bits, int BytesPerRow, int BitSkip, int Width, int Height)                           //
//                                                                                                                //
// Counts the number of black pixels in a rectangle which should be shrinked to one pixel.                        //
//                                                                                                                //
// u_char *Bits                         pointer to the top row of the rectangle in the bitmap                     //
// int    BytesPerRow                   bytes per row of the bitmap                                               //
// int    BitSkip                       left edge of the rectangle                                                //
// int    Width                         width of the rectangle                                                    //
// int    Height                        height of the rectangle                                                   //
//                                                                                                                //
// Result:                              number indicating color of the pixel                                      //
//                                                                                                                //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Glyph::Sample(u_char *Bits, int BytesPerRow, int BitSkip, int Width, int Height)
{
  static const char SampleCount[] =
  {
    0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6
  };

  u_char *ptr;
  u_char *EndPtr;
  u_char *cp;
  int    BitsLeft;
  int    n;
  int    BitShift;
  int    wid;

  ptr      = Bits + BitSkip / 8;
  EndPtr   = Bits + Height * BytesPerRow;
  BitsLeft = Width;
  BitShift = 8 - (BitSkip & 7);
  n        = 0;

  while(BitsLeft)
  {
    wid = BitShift;

    if(wid > BitsLeft)
      wid = BitsLeft;

    if(wid > 6)
      wid = 6;

    BitShift -= wid;

    for(cp = ptr; cp < EndPtr; cp = cp + BytesPerRow)
      n += SampleCount[(*cp >> BitShift) & BitMasks[wid]];

    if(BitShift == 0)
    {
      BitShift = 8;
      ptr++;
    }
    BitsLeft -= wid;
  }
  return n;
}
